{-# OPTIONS -cpp #-}
{-# OPTIONS -fglasgow-exts #-}

module Interface.Options
    ( Options(..)
    , Flag
    , parseStandardOptions
    , parsePluginOptions
    , defaultOptions
    , standardOptions_
    , isLiterate
    , mapFlag
    , usage
    ) where

import Prelude hiding (print, putStr, putStrLn)
import IO.Output

import Control.Monad.Error	( MonadError(catchError) )
import Data.List		( isSuffixOf )
import System.Console.GetOpt	(getOpt, usageInfo, ArgOrder(ReturnInOrder)
				, OptDescr(..), ArgDescr(..)
				)
import Utils.Monads		( readM )
import Utils.Files		( slash )

-- | This should probably go somewhere else.
isLiterate :: FilePath -> Bool
isLiterate file = ".lagda" `isSuffixOf` file

-- OptDescr is a Functor --------------------------------------------------

instance Functor OptDescr where
    fmap f (Option short long arg descr) = Option short long (fmap f arg) descr

instance Functor ArgDescr where
    fmap f (NoArg x)	= NoArg (f x)
    fmap f (ReqArg p s) = ReqArg (f . p) s
    fmap f (OptArg p s) = OptArg (f . p) s

data Options =
    Options { optInputFiles		:: [FilePath]
	    , optIncludeDirs		:: [FilePath]
	    , optCompile		:: Bool
	    , optOutputFile		:: Maybe FilePath
	    , optShowVersion		:: Bool
	    , optTexify			:: Bool
	    , optGenVimFile		:: Bool
	    , optGenEmacsFile		:: Bool
	    , optShowHelp		:: Bool
	    , optInteractive		:: Bool
	    , optVerbose		:: Int
	    , optCompletenessDepth	:: Int
	    }
    deriving Show

-- | Map a function over the long options. Also removes the short options.
--   Used to add the plugin name to the plugin options (see 'Plugins.Plugin').
mapFlag :: (String -> String) -> OptDescr a -> OptDescr a
mapFlag f (Option _ long arg descr) = Option [] (map f long) arg descr

defaultOptions :: Options
defaultOptions =
    Options { optInputFiles		= []
	    , optIncludeDirs		= [_ROOT_ ++ slash:"lib", _ROOT_ ++ slash:"cover"]
	    , optOutputFile		= Nothing
	    , optTexify			= False
	    , optCompile		= False
	    , optGenVimFile		= False
	    , optGenEmacsFile		= False
	    , optShowVersion		= False
	    , optShowHelp		= False
	    , optInteractive		= False
	    , optVerbose		= 1
	    , optCompletenessDepth	= 2
	    }

{- | @f :: Flag opts@  is an action on the option record that results from
parsing an option.  @f opts@ produces either an error message or an updated options record -}
type Flag opts	= opts -> Either String opts

outputFlag f o	    = return $ o { optOutputFile    = Just f		     }
texifyFlag o	    = return $ o { optTexify = True }
genVimFlag o	    = return $ o { optGenVimFile = True }
genEmacsFlag o	    = return $ o { optGenEmacsFile = True }
compileFlag o
    | optInteractive o	= fail "cannot use --compile with --interactive"
    | otherwise		= return $ o { optCompile = True }
inputFlag f o	    = return $ o { optInputFiles    = optInputFiles o ++ [f] }
versionFlag o	    = return $ o { optShowVersion   = True		     }
helpFlag o	    = return $ o { optShowHelp	    = True		     }
interactiveFlag o
    | optCompile o  = fail "cannot use --interactive with --compile"
    | otherwise	    = return $ o { optInteractive   = True }
includeFlag d o	    = return $ o { optIncludeDirs   = d : optIncludeDirs o   }
verboseFlag s o	    =
    do	n <- readM s
	return o { optVerbose = n }
    `catchError`
	\_ -> fail "option '--verbose' requires an integer argument"
completenessFlag s o =
    do	d <- readM s
	return o { optCompletenessDepth = d }
    `catchError`
	\_ -> fail "option '--completeness-depth requires an integer argument"

standardOptions :: [OptDescr (Flag Options)]
standardOptions =
    [ Option ['V']  ["version"]	(NoArg versionFlag) "show version number"
    , Option ['?']  ["help"]	(NoArg helpFlag)    "show this help"
    , Option ['v']  ["verbose"]	(ReqArg verboseFlag "N")
		    "set verbosity level to N"
    , Option ['i']  ["include-path"] (ReqArg includeFlag "DIR")
		    "look for imports in DIR"
    , Option ['c']  ["compile"]	(NoArg compileFlag) "compile to Haskell"
    , Option ['t']  ["texify"] (NoArg texifyFlag)
		    "generate LaTeX"
    , Option []	    ["vim"] (NoArg genVimFlag) "generate vim highlight file"
    , Option []	    ["emacs"] (NoArg genEmacsFlag) "generate Emacs highlight file"
    , Option ['o']  ["output"] (ReqArg outputFlag "FILE")
		    "write output to FILE (only with -c or -t)"
    , Option ['I']  ["interactive"] (NoArg interactiveFlag)
		    "start in interactive mode"
    , Option []	    ["completeness-depth"] (ReqArg completenessFlag "DEPTH")
		    "do DEPTH case distinctions when deciding if a type is empty (default 2)"
    ]

-- | Used for printing usage info.
standardOptions_ :: [OptDescr ()]
standardOptions_ = map (fmap $ const ()) standardOptions

-- | Don't export
parseOptions' ::
    String -> [String] -> opts ->
    [OptDescr (Flag opts)] -> (String -> Flag opts) -> Either String opts
parseOptions' progName argv defaults opts fileArg =
    case (getOpt (ReturnInOrder fileArg) opts argv) of
	(o,_,[])    -> foldl (>>=) (return defaults) o
	(_,_,errs)  -> fail $ concat errs

-- | Parse the standard options.
parseStandardOptions :: String -> [String] -> Either String Options
parseStandardOptions progName argv =
    parseOptions' progName argv defaultOptions standardOptions inputFlag

-- | Parse options for a plugin.
parsePluginOptions ::
    String -> [String] ->
    opts -> [OptDescr (Flag opts)] ->
    Either String opts
parsePluginOptions progName argv defaults opts =
    parseOptions'
	progName argv defaults opts
	(\s _ -> fail $ "Internal error: Flag " ++ s ++ " passed to a plugin")

-- | The usage info message. The argument is the program name (probably
--   agdaLight).
usage :: [OptDescr ()] -> [(String, String, [String], [OptDescr ()])] -> String -> String
usage options pluginInfos progName =
	usageInfo (header progName) options ++
	"\nPlugins:\n" ++
        indent (concatMap pluginMsg pluginInfos)
    
    where
	header progName = unlines [ "agdaLight version " ++ _VERSION_ ++
				    " in " ++ _ROOT_
				  , ""
				  , "Usage: " ++ progName ++ " [OPTIONS...] FILE"
				  ]

	indent = unlines . map ("  " ++) . lines

        pluginMsg (name, help, inherited, opts) 
            | null opts && null inherited = optHeader
            | otherwise = usageInfo (optHeader ++
                                     "  Plugin-specific options:" ++
				     inheritedOptions inherited
				     ) opts
            where
		optHeader = "\n" ++ name ++ "-plugin:\n" ++ indent help
		inheritedOptions [] = ""
		inheritedOptions pls =
		    "\n    Inherits options from: " ++ unwords pls

